# ==================================================================================================
# Copyright 2015 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================

"""
get percentile for values (https://en.wikipedia.org/wiki/Percentile)
(roughly Linear Interpolation Between Closest Ranks)

From:
  https://github.com/twitter/zktraffic/blob/master/zktraffic/stats/util.py
"""

import math


def percentile(values, percent, key=lambda k: k):

  assert isinstance(values, (list, tuple))

  idx = (len(values) - 1) * percent
  floor = math.floor(idx)
  ceil = math.ceil(idx)
  if floor == ceil:
    return key(values[int(idx)])

  a = key(values[int(floor)]) * (ceil - idx)
  b = key(values[int(ceil)]) * (idx - floor)

  return a + b




from __future__ import print_function

from collections import deque
from threading import Lock, Thread

import logging
import sys
import time
import traceback

from util import get_ip, get_ip_packet

from scapy.sendrecv import sniff
from scapy.config import conf as scapy_conf


scapy_conf.logLevel = logging.ERROR  # shush scappy


class Stream(object):
    """ A representation of a TCP stream. """
    def __init__(self, src, dst):
        self._packets = []
        self._src = src
        self._dst = dst
        self._length = 0
        self._remaining = 0
        self._next_seq_id = -1
        self._lock_packets = Lock()

    def __str__(self):
        return '%s<->%s (length: %d, remaining: %d, seq_id: %d)' % (
            self.src, self.dst, self.length, self.remaining, self._next_seq_id)

    @property
    def length(self):
        """ how many bytes have been through the stream """
        return self._length

    @property
    def remaining(self):
        return self._remaining

    @property
    def src(self):
        return self._src

    @property
    def dst(self):
        return self._dst

    def pop(self, nbytes):
        """ pops packets with _at least_ nbytes of payload """
        size = 0
        popped = []
        with self._lock_packets:
            while size < nbytes:
                try:
                    packet = self._packets.pop(0)
                    size += len(packet.data.data)
                    self._remaining -= len(packet.data.data)
                    popped.append(packet)
                except IndexError:
                    break
        return popped

    def pop_data(self, nbytes):
        """ similar to pop, but returns payload + last timestamp """
        last_timestamp = 0
        data = []
        for packet in self.pop(nbytes):
            last_timestamp = packet.timestamp
            data.append(packet.data.data)

        return ''.join(data), last_timestamp

    def push(self, ip_packet):
        """ push the packet into the queue """

        data_len = len(ip_packet.data.data)
        seq_id = ip_packet.data.seq

        if data_len == 0:
            self._next_seq_id = seq_id
            return False

        # have we seen this packet?
        if self._next_seq_id != -1 and seq_id != self._next_seq_id:
            return False

        self._next_seq_id = seq_id + data_len

        with self._lock_packets:
            # Note: we only account for payload (i.e.: tcp data)
            self._length += len(ip_packet.data.data)
            self._remaining += len(ip_packet.data.data)

            self._packets.append(ip_packet)

        return True


class Dispatcher(Thread):
    """Dispatches streams to handlers """
    def __init__(self, packet_queue):
        super(Dispatcher, self).__init__()
        self.setDaemon(True)
        self._queue = packet_queue
        self._streams = {}
        self._handlers = []
        self.start()

    @property
    def empty(self):
        return len(self._queue) == 0

    def add_handler(self, stream_handler):
        if stream_handler is None:
            return

        if stream_handler in self._handlers:
            raise ValueError('handler already registered')

        self._handlers.append(stream_handler)

    def run(self, *args, **kwargs):
        """ Deal with the incoming packets """
        while True:
            try:
                timestamp, ip_p = self._queue.popleft()

                src_ip = get_ip(ip_p, ip_p.src)
                dst_ip = get_ip(ip_p, ip_p.dst)

                src = intern('%s:%s' % (src_ip, ip_p.data.sport))
                dst = intern('%s:%s' % (dst_ip, ip_p.data.dport))
                key = intern('%s<->%s' % (src, dst))

                stream = self._streams.get(key)
                if stream is None:
                    stream = Stream(src, dst)
                    self._streams[key] = stream

                # HACK: save the timestamp
                setattr(ip_p, 'timestamp', timestamp)
                pushed = stream.push(ip_p)

                if not pushed:
                    continue

                # let listeners know about the updated stream
                for handler in self._handlers:
                    try:
                        handler(stream)
                    except Exception as ex:
                        print('handler exception: %s' % ex)
            except Exception:
                time.sleep(0.00001)


class Sniffer(Thread):
    """ A generic & simple packet sniffer """

    def __init__(self, iface, port, stream_handler=None, offline=None, ip=None):
        """A Sniffer that merges packets into a stream

        Params:
            ``iface``           The interface in which to listen
            ``port``            The TCP port that we care about
            ``stream_handler``  The callback for each stream
            ``offline``         Path to a pcap file
            ``ip``              A list of IPs that we care about
        """
        super(Sniffer, self).__init__()
        self.setDaemon(True)

        self._iface = iface
        self._port = port
        self._offline = offline
        self._ip = ip if ip else []
        self._queue = deque()  # TODO: maxlen?
        self._dispatcher = Dispatcher(self._queue)

        self._dispatcher.add_handler(stream_handler)

        self._wants_stop = False

        self.start()

    @property
    def dispatcher(self):
        return self._dispatcher

    @property
    def pending_ip_packets(self):
        return len(self._queue)

    def add_handler(self, stream_handler):
        self._dispatcher.add_handler(stream_handler)

    def run(self):
        pfilter = 'port %d' % self._port
        try:
            kwargs = {
                'filter': pfilter,
                'store': 0,
                'prn': self._handle_packet,
                'iface': self._iface,
                'stop_filter': lambda p: self._wants_stop,
                }

            if self._offline:
                kwargs['offline'] = self._offline

            sniff(**kwargs)
        except Exception as ex:
            if 'Not a pcap capture file' in str(ex):
                print('%s is not a valid pcap file' % self._offline)
                return
            print('Error: %s: %s (device: %s)' % (ex, traceback.format_exc(), self._iface))
        finally:
            if self._offline:
                # drain dispatcher
                while not self._dispatcher.empty:
                    time.sleep(0.1)

    def stop(self, wait_for_stopped=False):
        if not self.isAlive():
            return

        self._wants_stop = True

        if wait_for_stopped:
            while self.isAlive():
                time.sleep(0.01)

    def _handle_packet(self, packet):
        try:
            ip_p = get_ip_packet(packet.load, 0, self._port)
        except ValueError:
            return

        ip_data = getattr(ip_p, 'data', None)
        if ip_data is None:
            return

        if ip_data.sport != self._port and ip_data.dport != self._port:
            return

        if self._ip:
            src_ip = get_ip(ip_p, ip_p.src)
            dst_ip = get_ip(ip_p, ip_p.dst)

            if src_ip not in self._ip and dst_ip not in self._ip:
                return

        self._queue.append((packet.time, ip_p))



""" helpers for deserializing Thrift messages """

from struct import unpack
from thrift.Thrift import TMessageType

from thrift.protocol.TBinaryProtocol import TBinaryProtocol
from thrift.protocol.TCompactProtocol import TCompactProtocol
from thrift.protocol.TJSONProtocol import TJSONProtocol
from thrift.transport import TTransport
from thrift_tools.thrift_struct import ThriftStruct


class ThriftMessage(object):
    def __init__(self, method, mtype, seqid, args, header=(), length=-1):
        self._method = method
        self._type = mtype
        self._seqid = seqid
        if not isinstance(args, ThriftStruct):
            raise ValueError('args must be a ThriftStruct instance')
        self._args = args
        self._header = header  # finagle-thrift prepends this to each call
        self._length = length

    def __len__(self):
        return self._length

    @property
    def method(self):
        return self._method

    @property
    def type(self):
        return self._type

    @property
    def seqid(self):
        return self._seqid

    @property
    def args(self):
        return self._args

    @property
    def header(self):
        return self._header

    def __str__(self):
        return 'method=%s, type=%s, seqid=%s, header=%s, fields=%s' % (
            self.method, self.type, self.seqid, self.header, str(self.args))

    @property
    def as_dict(self):
        return {
            'method': self.method,
            'type': self.type,
            'seqid': self.seqid,
            'header': self.header,
            'args': self.args,
            'length': len(self),
        }

    MAX_METHOD_LENGTH = 70

    # For Binary, this is i32 + str + i32
    # For Compact, the empty ping() gets through in 8 bytes
    MIN_MESSAGE_SIZE = 8

    # some sane defaults to keep memory usage tight
    MAX_FIELDS = 1000
    MAX_LIST_SIZE = 10000
    MAX_MAP_SIZE = 10000
    MAX_SET_SIZE = 10000

    @classmethod
    def read(cls, data,
             protocol=None,
             fallback_protocol=TBinaryProtocol,
             finagle_thrift=False,
             max_fields=MAX_FIELDS,
             max_list_size=MAX_LIST_SIZE,
             max_map_size=MAX_MAP_SIZE,
             max_set_size=MAX_SET_SIZE,
             read_values=False):
        """ tries to deserialize a message, might fail if data is missing """

        # do we have enough data?
        if len(data) < cls.MIN_MESSAGE_SIZE:
            raise ValueError('not enough data')

        if protocol is None:
            protocol = cls.detect_protocol(data, fallback_protocol)
        trans = TTransport.TMemoryBuffer(data)
        proto = protocol(trans)

        # finagle-thrift prepends a RequestHeader
        #
        # See: http://git.io/vsziG
        header = None
        if finagle_thrift:
            try:
                header = ThriftStruct.read(
                    proto,
                    max_fields,
                    max_list_size,
                    max_map_size,
                    max_set_size,
                    read_values)
            except:
                # reset stream, maybe it's not finagle-thrift
                trans = TTransport.TMemoryBuffer(data)
                proto = protocol(trans)

        # unpack the message
        method, mtype, seqid = proto.readMessageBegin()
        mtype = cls.message_type_to_str(mtype)

        if len(method) == 0 or method.isspace() or method.startswith(' '):
            raise ValueError('no method name')

        if len(method) > cls.MAX_METHOD_LENGTH:
            raise ValueError('method name too long')

        # we might have made it until this point by mere chance, so filter out
        # suspicious method names
        valid = range(33, 127)
        if any(ord(char) not in valid for char in method):
            raise ValueError('invalid method name' % method)

        args = ThriftStruct.read(
            proto,
            max_fields,
            max_list_size,
            max_map_size,
            max_set_size,
            read_values)

        proto.readMessageEnd()

        # Note: this is a bit fragile, the right thing would be to count bytes
        # as we read them (i.e.: when calling readI32, etc).
        msglen = trans._buffer.tell()

        return cls(method, mtype, seqid, args, header, msglen), msglen

    @classmethod
    def detect_protocol(cls, data, default=None):
        """ TODO: support fbthrift, finagle-thrift, finagle-mux, CORBA """
        if cls.is_compact_protocol(data):
            return TCompactProtocol
        elif cls.is_binary_protocol(data):
            return TBinaryProtocol
        elif cls.is_json_protocol(data):
            return TJSONProtocol

        if default is None:
            raise ValueError('Unknown protocol')

        return default

    COMPACT_PROTOCOL_ID = 0x82

    @classmethod
    def is_compact_protocol(cls, data):
        result, = unpack('!B', data[0])
        return result == cls.COMPACT_PROTOCOL_ID

    BINARY_PROTOCOL_VERSION_MASK = -65536  # 0xffff0000
    BINARY_PROTOCOL_VERSION_1 = -2147418112  # 0x80010000

    @classmethod
    def is_binary_protocol(cls, data):
        val, = unpack('!i', data[0:4])
        if val >= 0:
            return False
        version = val & cls.BINARY_PROTOCOL_VERSION_MASK
        return version == cls.BINARY_PROTOCOL_VERSION_1

    @classmethod
    def is_json_protocol(cls, data):
        # FIXME: more elaborate parsing would make this more robust
        return data.startswith('[1')

    @staticmethod
    def message_type_to_str(mtype):
        if mtype == TMessageType.CALL:
            return 'call'
        elif mtype == TMessageType.REPLY:
            return 'reply'
        elif mtype == TMessageType.EXCEPTION:
            return 'exception'
        elif mtype == TMessageType.ONEWAY:
            return 'oneway'
        else:
            raise ValueError('Unknown message type: %s' % mtype)


"""
    Diff thrift structures & messages

    Examples:
        > while not finished:
                msg_a, _ = ThriftMessage.read(data_stream_a, read_values=True)
                msg_b, _ = ThriftMessage.read(data_stream_b, read_values=True)
                args_diff = ThriftDiff.of_messages(msg_a, msg_b)
                for diff in args_diff:
                    print diff.common_fields()
        [(ThriftField(field_type=str, field_id=1, value="hello world"),
            ThriftField(field_type=str, field_id=1, value="bye world"))]

        > t_diff = ThriftDiff.of_structs(struct_a, struct_b)
        > print t_diff.fields_with_same_value()
        ThriftField(field_type=str, field_id=1, value="hello world")
    """


class ThriftDiff:

    def __init__(self, struct_a, struct_b):
        self._a = struct_a
        self._b = struct_b
        self._common_fields = []
        self._fields_only_in_a = []
        self._fields_only_in_b = []
        self._fields_with_different_value = []
        self._fields_with_same_value = []

    @classmethod
    def of_structs(cls, a, b):
        """
        Diff two thrift structs and return the result as a ThriftDiff instance
        """
        t_diff = ThriftDiff(a, b)
        t_diff._do_diff()
        return t_diff

    @classmethod
    def of_messages(cls, msg_a, msg_b):
        """
        Diff two thrift messages by comparing their args, raises exceptions if
        for some reason the messages can't be diffed. Only args of type 'struct'
        are compared.

        Returns a list of ThriftDiff results - one for each struct arg
        """
        ok_to_diff, reason = cls.can_diff(msg_a, msg_b)
        if not ok_to_diff:
            raise ValueError(reason)
        return [cls.of_structs(x.value, y.value)
                for x, y in zip(msg_a.args, msg_b.args)
                if x.field_type == 'struct']

    @staticmethod
    def can_diff(msg_a, msg_b):
        """
        Check if two thrift messages are diff ready.

        Returns a tuple of (boolean, reason_string), i.e. (False, reason_string)
        if the messages can not be diffed along with the reason and
        (True, None) for the opposite case
        """
        if msg_a.method != msg_b.method:
            return False, 'method name of messages do not match'
        if len(msg_a.args) != len(msg_b.args) \
                or not msg_a.args.is_isomorphic_to(msg_b.args):
            return False, 'argument signature of methods do not match'
        return True, None

    def _do_diff(self):
        self._common_fields = [(x, y)
                               for x, y in zip(self._a.fields, self._b.fields)
                               if x.is_isomorphic_to(y)]
        self._fields_only_in_a = self._unique_fields(self._common_fields,
                                                     self._a.fields)
        self._fields_only_in_b = self._unique_fields(self._common_fields,
                                                     self._b.fields)
        # Go over the common fields and filter them into two sets, one for
        # the case when field values match and other for the ones where they
        # don't match
        for field_pair in self._common_fields:
            if field_pair[0] == field_pair[1]:
                # For common fields with same values, just record one of the
                # fields in the pair
                self._fields_with_same_value.append(field_pair[0])
            else:
                self._fields_with_different_value.append(field_pair)

    @staticmethod
    def _unique_fields(common_fields, all_fields):
        unique_fields = []
        for x in all_fields:
            is_unique = True
            for field_pair in common_fields:
                if x.is_isomorphic_to(field_pair[0]):
                    is_unique = False
                    break
            if is_unique:
                unique_fields.append(x)
        return unique_fields

    @property
    def common_fields(self):
        """
        List of isomorphically equivalent field pairs which may or may not have
        same value
        """
        return self._common_fields

    @property
    def fields_only_in_a(self):
        """
        List of fields exclusive to first struct
        """
        return self._fields_only_in_a

    @property
    def fields_only_in_b(self):
        """
        List of fields exclusive to second struct
        """
        return self._fields_only_in_b

    @property
    def fields_with_same_value(self):
        """
        List of isomorphically equivalent fields for which value is also equal
        Note: this doesn't return a list of 'pairs'
        """
        return self._fields_with_same_value

    @property
    def field_with_different_value(self):
        """
        List of isomorphically equivalent field pairs for which value is NOT
        equal
        """
        return self._fields_with_different_value


""" A StreamHandler that extracts messages from streams """

from __future__ import print_function

from collections import defaultdict

import sys
import traceback

from .thrift_message import ThriftMessage


class StreamContext(object):
    def __init__(self):
        self.bytes = ''


class StreamHandler(object):
    def __init__(self,
                 outqueue,
                 protocol=None,
                 finagle_thrift=False,
                 max_message_size=1024*1000,
                 read_values=False,
                 debug=False):
        self._contexts_by_streams = defaultdict(StreamContext)
        self._pop_size = 1024  # TODO: what's a good value here?
        self._outqueue = outqueue
        self._protocol = protocol
        self._finagle_thrift = finagle_thrift
        self._max_message_size = max_message_size
        self._debug = debug
        self._read_values = read_values
        self._seen_messages = 0
        self._recognized_streams = set()  # streams from which msgs have been read

    def __call__(self, *args, **kwargs):
        self.handler(*args, **kwargs)

    @property
    def seen_streams(self):
        return len(self._contexts_by_streams)

    @property
    def recognized_streams(self):
        return len(self._recognized_streams)

    @property
    def unrecognized_streams(self):
        return self.seen_streams - self.recognized_streams

    @property
    def pending_thrift_msgs(self):
        return len(self._outqueue)

    @property
    def seen_thrift_msgs(self):
        return self._seen_messages

    def handler(self, stream):
        context = self._contexts_by_streams[stream]
        bytes, timestamp = stream.pop_data(self._pop_size)
        context.bytes += bytes

        # EMSGSIZE
        if len(context.bytes) >= self._max_message_size:
            if self._debug:
                print('Dropping bytes, dropped size: %d' % len(context.bytes))
            context.bytes = ''
            return

        # FIXME: a bit of brute force to find the start of a message.
        #        Is there a magic byte/string we can look for?

        view = memoryview(context.bytes)
        for idx in range(0, len(context.bytes)):
            try:
                data_slice = view[idx:].tobytes()
                msg, msglen = ThriftMessage.read(
                    data_slice,
                    protocol=self._protocol,
                    finagle_thrift=self._finagle_thrift,
                    read_values=self._read_values)
            except EOFError:
                continue
            except Exception as ex:
                if self._debug:
                    print('Bad message for stream %s: %s: %s\n(idx=%d) '
                          '(context size=%d)' % (
                            stream,
                            ex,
                            traceback.format_exc(),
                            idx,
                            len(context.bytes)),
                          file=sys.stderr
                          )
                continue

            self._recognized_streams.add(stream)
            self._seen_messages += 1
            self._outqueue.append((timestamp, stream.src, stream.dst, msg))
            context.bytes = context.bytes[idx + msglen:]
            break



import unittest
from thrift_tools.thrift_struct import ThriftField, ThriftStruct


class ThriftStructTestCase(unittest.TestCase):
    """ Unit tests for ThriftStruct"""

    def setUp(self):
        self.monday = ThriftField("string", 1, "monday")
        self.tuesday = ThriftField("string", 1, "tuesday")
        self.todo_1 = ThriftField("string", 2, "run")
        self.todo_2 = ThriftField("string", 2, "swim")
        self.weekday = ThriftField("bool", 1, True)
        self.calendar_1 = ThriftField("struct", 1,
                                      ThriftStruct((self.monday, self.todo_1)))
        self.calendar_2 = ThriftField("struct", 1,
                                      ThriftStruct((self.tuesday, self.todo_2)))
        self.calendar_3 = ThriftField("struct", 1,
                                      ThriftStruct((self.monday, self.todo_1)))
        self.calendar_4 = ThriftField("struct", 2,
                                      ThriftStruct((self.monday, self.todo_1)))

    def tearDown(self):
        pass

    def test_is_isomorphic_to(self):
        struct_1 = self.calendar_1.value
        struct_2 = self.calendar_2.value
        struct_3 = ThriftStruct([self.monday])
        self.assertTrue(struct_1.is_isomorphic_to(struct_2))
        self.assertFalse(struct_1.is_isomorphic_to(struct_3))


class ThriftFieldTestCase(ThriftStructTestCase):
    """ Unit tests for ThriftField"""

    def test_is_isomorphic_to(self):
        # Test primitive fields
        self.assertTrue(self.monday.is_isomorphic_to(self.tuesday))
        self.assertFalse(self.monday.is_isomorphic_to(self.todo_1))
        self.assertFalse(self.monday.is_isomorphic_to(self.weekday))
        # Tests struct fields
        self.assertTrue(self.calendar_1.is_isomorphic_to(self.calendar_2))
        self.assertFalse(self.calendar_1.is_isomorphic_to(self.monday))

    def test_eq(self):
        self.assertFalse(self.monday == self.tuesday)
        self.assertFalse(self.monday == self.tuesday)
        self.assertTrue(self.calendar_1 == self.calendar_3)
        self.assertFalse(self.calendar_1 == self.calendar_4)


from collections import deque

import unittest

from thrift_tools.sniffer import Sniffer
from thrift_tools.stream_handler import StreamHandler
from thrift_tools.thrift_struct import ThriftField, ThriftStruct

from .util import get_pcap_path


class BasicTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_calculator_service_binary(self):
        self._test_protocol('binary')

    def test_calculator_service_compact(self):
        self._test_protocol('compact')

    def test_calculator_service_json(self):
        self._test_protocol('json')

    def test_finagle(self):
        queue = deque()
        pcap_file = get_pcap_path('finagle-thrift')
        handler = StreamHandler(queue, read_values=True, finagle_thrift=True)

        sniffer = Sniffer('ignore', 9090, handler, offline=pcap_file)
        sniffer.join()

        self.assertEquals(len(queue), 22)

        # is this finagle-thrift indeed?
        _, src, dst, msg = queue.popleft()
        self.assertEquals(msg.method, '__can__finagle__trace__v3__')
        self.assertEquals(msg.type, 'call')
        self.assertEquals(len(msg.args), 0)

        _, src, dst, msg = queue.popleft()
        self.assertEquals(msg.method, '__can__finagle__trace__v3__')
        self.assertEquals(msg.type, 'reply')
        self.assertEquals(len(msg.args), 0)

        # the search() call
        _, src, dst, msg = queue.popleft()
        self.assertEquals(msg.method, 'search')
        self.assertEquals(msg.type, 'call')

        # inspect the header & the contexts
        self.assertEquals(len(msg.header), 4)
        self.assertEquals(msg.header[0], ThriftField('i64', 1, -8277104800942727271))
        self.assertEquals(msg.header[1], ThriftField('i64', 2, -8277104800942727271))
        self.assertEquals(msg.header[2], ThriftField('i64', 7, 0))

        contexts = msg.header[3].value
        self.assertEquals(contexts[0][0].value,
                          'com.twitter.finagle.tracing.TraceContext')
        self.assertEquals(contexts[1][0].value,
                          'com.twitter.finagle.Deadline')

        self.assertEquals(msg.args, ThriftStruct([ThriftField('string', 1, 'foo')]))

        # the reply
        _, src, dst, msg = queue.popleft()
        self.assertEquals(msg.method, 'search')
        self.assertEquals(msg.type, 'reply')
        self.assertEquals(msg.args, ThriftStruct([ThriftField('list', 0, ['one', 'two', 'three'])]))

    def _test_protocol(self, protoname):
        queue = deque()
        pcap_file = get_pcap_path('calc-service-%s' % protoname)
        handler = StreamHandler(queue, read_values=True, debug=True)

        sniffer = Sniffer('ignore', 9090, handler, offline=pcap_file)
        sniffer.join()

        self.assertEquals(len(queue), 10)

        # the ping call
        _, src, dst, msg = queue.popleft()
        self.assertEquals(msg.method, 'ping')
        self.assertEquals(msg.type, 'call')
        self.assertEquals(len(msg.args), 0)

        _, src, dst, msg = queue.popleft()
        self.assertEquals(msg.method, 'ping')
        self.assertEquals(msg.type, 'reply')
        self.assertEquals(len(msg.args), 0)

        # a succesful add
        _, src, dst, msg = queue.popleft()
        self.assertEquals(msg.method, 'add')
        self.assertEquals(msg.type, 'call')
        self.assertEquals(len(msg.args), 2)
        self.assertEquals(msg.args[0], ThriftField('i32', 1, 1))
        self.assertEquals(msg.args[1], ThriftField('i32', 2, 1))

        _, src, dst, msg = queue.popleft()
        self.assertEquals(msg.method, 'add')
        self.assertEquals(msg.type, 'reply')
        self.assertEquals(len(msg.args), 1)
        self.assertEquals(msg.args[0], ThriftField('i32', 0, 2))

        # a failed calculate call
        _, src, dst, msg = queue.popleft()
        self.assertEquals(msg.method, 'calculate')
        self.assertEquals(msg.type, 'call')
        self.assertEquals(len(msg.args), 2)
        self.assertEquals(msg.args[0], ThriftField('i32', 1, 1))
        self.assertEquals(
            msg.args[1],
            ThriftField('struct', 2,
                        ThriftStruct(
                            [ThriftField('i32', 1, 1),
                             ThriftField('i32', 2, 0),
                             ThriftField('i32', 3, 4)])))
        _, src, dst, msg = queue.popleft()
        self.assertEquals(msg.method, 'calculate')
        self.assertEquals(msg.type, 'reply')
        self.assertEquals(len(msg.args), 1)
        self.assertEquals(
            msg.args[0],
            ThriftField('struct', 1,
                        ThriftStruct([ThriftField('i32', 1, 4),
                                      ThriftField('string', 2, 'Cannot divide by 0')])))

        # a successful calculate call
        _, src, dst, msg = queue.popleft()
        self.assertEquals(msg.method, 'calculate')
        self.assertEquals(msg.type, 'call')
        self.assertEquals(len(msg.args), 2)
        self.assertEquals(
            msg.args[1],
            ThriftField('struct', 2, ThriftStruct([ThriftField('i32', 1, 15),
                                                   ThriftField('i32', 2, 10),
                                                   ThriftField('i32', 3, 2)])))

        _, src, dst, msg = queue.popleft()
        self.assertEquals(msg.method, 'calculate')
        self.assertEquals(msg.type, 'reply')
        self.assertEquals(len(msg.args), 1)
        self.assertEquals(msg.args[0], ThriftField('i32', 0, 5))

        # getStruct
        _, src, dst, msg = queue.popleft()
        self.assertEquals(msg.method, 'getStruct')
        self.assertEquals(msg.type, 'call')
        self.assertEquals(len(msg.args), 1)
        self.assertEquals(msg.args[0], ThriftField('i32', 1, 1))

        _, src, dst, msg = queue.popleft()
        self.assertEquals(msg.method, 'getStruct')
        self.assertEquals(msg.type, 'reply')
        self.assertEquals(len(msg.args), 1)
        self.assertEquals(
            msg.args[0],
            ThriftField('struct', 0, ThriftStruct([ThriftField('i32', 1, 1),
                                                   ThriftField('string', 2, '5')])))



from StringIO import StringIO

import unittest

from thrift.protocol.TBinaryProtocol import TBinaryProtocol
from thrift_tools.message_sniffer import MessageSnifferOptions, MessageSniffer
from thrift_tools.printer import (
    FormatOptions, LatencyPrinter, PairedPrinter, Printer)

from .util import get_pcap_path


def options():
    return MessageSnifferOptions(
        iface='ignore',
        port=9090,
        ip=None,
        pcap_file=get_pcap_path('calc-service-binary'),
        protocol=TBinaryProtocol,
        finagle_thrift=False,
        read_values=True,
        max_queued=2000,
        max_message_size=2000,
        debug=False)


def format_options():
    return FormatOptions(
        pretty_printer=False,
        is_color=False,
        show_header=False,
        show_fields=False,
        json=False,
        )


class PrinterTestCase(unittest.TestCase):
    def setUp(self):
        pass

    def tearDown(self):
        pass

    def test_default(self):
        output = StringIO()
        printer = Printer(format_options(), output)

        message_sniffer = MessageSniffer(options(), printer)
        message_sniffer.join()

        self._assertMessages(output)

    def test_paired(self):
        output = StringIO()
        printer = PairedPrinter(format_options(), output)

        message_sniffer = MessageSniffer(options(), printer)
        message_sniffer.join()

        self._assertMessages(output)

    def test_latency_printer(self):
        output = StringIO()
        printer = LatencyPrinter(expected_calls=3, output=output)

        message_sniffer = MessageSniffer(options(), printer)
        message_sniffer.join()

        self.assertIn(
            'method       count          avg          min          max          p90          p95          p99         p999',
            output.getvalue())
        self.assertIn(
            'ping             1  0.000156164  0.000156164  0.000156164  0.000156164  0.000156164  0.000156164  0.000156164',
            output.getvalue())
        self.assertIn(
            'calculate        1  0.000144005  0.000144005  0.000144005  0.000144005  0.000144005  0.000144005  0.000144005',
            output.getvalue())
        self.assertIn(
            'add              1  0.000103951  0.000103951  0.000103951  0.000103951  0.000103951  0.000103951  0.000103951',
            output.getvalue())

    def _assertMessages(self, output):
        """ this is a bit fragile... """
        self.assertIn(
            '[22:55:50:387214] 127.0.0.1:51112 -> 127.0.0.1:9090: method=ping, type=call, seqid=0',
            output.getvalue())
        self.assertIn(
            '[22:55:50:387370] 127.0.0.1:9090 -> 127.0.0.1:51112: method=ping, type=reply, seqid=0',
            output.getvalue())
        self.assertIn(
            '[22:55:50:387492] 127.0.0.1:51112 -> 127.0.0.1:9090: method=add, type=call, seqid=0',
            output.getvalue())
        self.assertIn(
            '[22:55:50:387596] 127.0.0.1:9090 -> 127.0.0.1:51112: method=add, type=reply, seqid=0',
            output.getvalue())
        self.assertIn(
            '[22:55:50:387696] 127.0.0.1:51112 -> 127.0.0.1:9090: method=calculate, type=call, seqid=0',
            output.getvalue())
        self.assertIn(
            '[22:55:50:387840] 127.0.0.1:9090 -> 127.0.0.1:51112: method=calculate, type=reply, seqid=0',
            output.getvalue())
        self.assertIn(
            '[22:55:50:388615] 127.0.0.1:51112 -> 127.0.0.1:9090: method=calculate, type=call, seqid=0',
            output.getvalue())
        self.assertIn(
            '[22:55:50:388725] 127.0.0.1:9090 -> 127.0.0.1:51112: method=calculate, type=reply, seqid=0',
            output.getvalue())
        self.assertIn(
            '[22:55:50:388811] 127.0.0.1:51112 -> 127.0.0.1:9090: method=getStruct, type=call, seqid=0',
            output.getvalue())
        self.assertIn(
            '[22:55:50:388905] 127.0.0.1:9090 -> 127.0.0.1:51112: method=getStruct, type=reply, seqid=0\n',
            output.getvalue())



from __future__ import print_function

from collections import deque, namedtuple
from threading import Thread

import time

from .sniffer import Sniffer
from .stream_handler import StreamHandler


MessageSnifferOptions = namedtuple('MessageSnifferOptions', [
    'iface',
    'port',
    'ip',
    'pcap_file',
    'protocol',
    'finagle_thrift',
    'read_values',
    'max_queued',
    'max_message_size',
    'debug',
])


STOP_MESSAGE = object()
DONE_TUPLE = (None, None, None, STOP_MESSAGE)


class MessageSniffer(Thread):
    def __init__(self, options, handler=None):
        self._options = options
        self._handlers = []
        self._queue = deque(maxlen=options.max_queued)

        self._handler = StreamHandler(
            self._queue,
            protocol=options.protocol,
            finagle_thrift=options.finagle_thrift,
            max_message_size=options.max_message_size,
            read_values=options.read_values,
            debug=options.debug)

        self._sniffer = Sniffer(
            options.iface, options.port,
            stream_handler=self._handler,
            ip=options.ip,
            offline=options.pcap_file)

        self.add_handler(handler)

        super(MessageSniffer, self).__init__()
        self.setDaemon(True)
        self.start()

    def status(self):
        return """
alive:                  %s
queue size:             %d
seen streams:           %d
unrecognized streams:   %d
seen thrift msgs:       %d
pending thrift msgs:    %d
sniffer alive:          %s
pending ip packets:     %d
dispatcher alive:       %s
""" % (self.isAlive(),
       len(self._queue),
       self._handler.seen_streams,
       self._handler.unrecognized_streams,
       self._handler.seen_thrift_msgs,
       self._handler.pending_thrift_msgs,
       self._sniffer.isAlive(),
       self._sniffer.pending_ip_packets,
       self._sniffer.dispatcher.isAlive()
       )

    def add_handler(self, handler):
        if handler is None:
            return

        if handler in self._handlers:
            raise ValueError('handler already registered')

        self._handlers.append(handler)

    def run(self):
        opts = self._options
        handler = self._handler
        sniffer = self._sniffer

        # main loop
        running = True
        while running:

            # if the sniffer finished and the queue is empty, we are done
            if not sniffer.isAlive() and len(self._queue) == 0:
                break

            try:
                timestamp, src, dst, msg = self._queue.popleft()
            except IndexError:
                time.sleep(0.001)
                continue

            if msg == STOP_MESSAGE:
                break

            # dispatch thrift messages to handlers
            for handler in self._handlers:
                try:
                    rv = handler(timestamp, src, dst, msg)
                    if not rv:
                        running = False
                        break
                except Exception as ex:
                    print('handler exception: %s' % ex)

        # leaving...
        if len(self._queue):
            print('%d messages left in the queue' % len(self._queue))

    def stop(self, wait_for_stopped=False):
        if not self.isAlive():
            return

        self._queue.append(DONE_TUPLE)

        if wait_for_stopped:
            while self.isAlive():
                time.sleep(0.01)



from collections import defaultdict, deque, namedtuple
from datetime import datetime

import json
import pprint
import sys
import time

from tabulate import tabulate

import colors

from .stats import percentile


fromtimestamp = datetime.fromtimestamp


FormatOptions = namedtuple('FormatOptions', [
    'pretty_printer',
    'is_color',
    'show_header',
    'show_fields',
    'json',
])


def print_color(s, color_id, output=sys.stdout):
    attr = colors.COLORS[color_id % len(colors.COLORS)]
    cfunc = getattr(colors, attr)
    output.write(cfunc(s))
    output.flush()


def print_msg(timestamp, src, dst, msg, format_opts,
              prefix='', indent=0, output=sys.stdout):
    timestr = fromtimestamp(timestamp).strftime('%H:%M:%S:%f')

    def pretty(part):
        if format_opts.pretty_printer:
            pp = pprint.PrettyPrinter(indent=4)
            part = pp.pformat(part)
        return part

    indent = ' ' * indent if indent else ''

    if format_opts.show_header:
        header_line = '%sheader: %s\n' % (indent, pretty(msg.header))
    else:
        header_line = ''

    if format_opts.show_fields:
        fields_line = '%sfields: %s\n' % (indent, pretty(msg.args))
    else:
        fields_line = ''

    if format_opts.json:
        parts = {
            'time': timestr,
            'src': src,
            'dst': dst,
            'method': msg.method,
            'type': msg.type,
            'seqid': msg.seqid,
            }

        if format_opts.show_header:
            parts['header'] = msg.header

        if format_opts.show_fields:
            parts['fields'] = msg.args

        outputstr = json.dumps(parts, indent=4) + '\n'
    else:
        outputstr = '%s[%s] %s -> %s: method=%s, type=%s, seqid=%d\n%s%s' % (
            prefix, timestr, src, dst, msg.method, msg.type, msg.seqid,
            header_line, fields_line)

    if format_opts.is_color:
        print_color(outputstr, src.__hash__())
    else:
        output.write(outputstr)
        output.flush()


class Printer(object):
    """ A simple message printer """

    def __init__(self, format_opts, output=sys.stdout):
        self._format_opts = format_opts
        self._output = output

    def __call__(self, timestamp, src, dst, msg):
        print_msg(timestamp, src, dst, msg, self._format_opts,
                  output=self._output)
        return True  # keep the sniffer running


class PairedPrinter(object):
    """ Pairs each request with its reply """
    def __init__(self, format_opts, output=sys.stdout):
        self._format_opts = format_opts
        self._output = output

        # msgs by [src][dst][method_name]
        self._requests = defaultdict(
            lambda: defaultdict(lambda: defaultdict(deque)))
        # ditto
        self._replies = defaultdict(
            lambda: defaultdict(lambda: defaultdict(deque)))

    def __call__(self, timestamp, src, dst, msg):
        """
        We need to match up each (request, reply) pair. Presumably,
        pcap _shouldn't_ deliver packets out of order, but
        things could get mixed up somewhere withing the
        TCP stream being reassembled and the StreamHandler
        thread. So, we don't assume that a 'reply' implies
        the corresponding 'call' has been seen.

        It could also be that we started sniffing after
        the 'call' message... but there's no easy way to tell
        (given we don't keep the startup time around...)
        """
        if msg.type == 'call':
            replies = self._replies[dst][src][msg.method]
            if len(replies) > 0:
                reply_timestamp, reply = replies.popleft()
                self._print_pair(
                    timestamp, msg, reply_timestamp, reply, src, dst)
            else:
                self._requests[src][dst][msg.method].append(
                    (timestamp, msg))
        elif msg.type == 'reply':
            requests = self._requests[dst][src][msg.method]
            if len(requests) > 0:
                request_timestamp, request = requests.popleft()
                self._print_pair(
                    request_timestamp, request, timestamp, msg,
                    dst, src)
            else:
                self._replies[src][dst][msg.method].append(
                    (timestamp, msg))
        else:
            print_msg(timestamp, src, dst, msg, self._format_opts,
                      output=self._output)

        return True  # keep the sniffer running

    def _print_pair(self, reqtime, request, reptime, reply, src, dst):
        print_msg(reqtime, src, dst, request, self._format_opts,
                  output=self._output)
        print_msg(reptime, dst, src, reply, self._format_opts,
                  prefix='------>', indent=8, output=self._output)


class LatencyPrinter(object):
    """ Reports latencies for the seen (req, rep) pairs """
    def __init__(self, expected_calls, output=sys.stdout):
         # msgs by [src][dst][method_name][seqid]
        self._requests = defaultdict(
            lambda: defaultdict(lambda: defaultdict(dict)))

        self._expected = expected_calls
        self._output = output

        self._seen = 0
        self._pairs = []
        self._latencies_by_method = defaultdict(list)

    def __call__(self, timestamp, src, dst, msg):
        """
        Slightly simplified logic wrt what PairedPrinter has:
        we assume 'call' messages will be seen before their
        corresponding 'reply'.
        """
        if msg.type == 'call':
            self._requests[src][dst][msg.method][msg.seqid] = (timestamp, msg)
        elif msg.type == 'reply':
            # have we seen the corresponding call?
            calls = self._requests[dst][src][msg.method]
            if msg.seqid in calls:
                request_timestamp, request = calls.pop(msg.seqid)
                latency = timestamp - request_timestamp
                self._latencies_by_method[msg.method].append(latency)
                self._seen += 1

                # let the user know we are still working
                self._output.write('\rCollecting (%d/%d)' % (
                        self._seen, self._expected))
                self._output.flush()

        if self._seen < self._expected:
            return True  # we still need more messages

        self.report()

        # what about unmatched calls?
        unmatched = 0
        for src, dst in self._requests.items():
            for method, calls in dst.items():
                unmatched += len(calls)

        if unmatched > 0:
            self._output.write('%d unmatched calls\n' % unmatched)

        return False  # we are done

    def report(self):
        """ get stats & show them """
        self._output.write('\r')

        sort_by = 'avg'
        results = {}
        for key, latencies in self._latencies_by_method.items():
            result = {}
            result['count'] = len(latencies)
            result['avg'] = sum(latencies) / len(latencies)
            result['min'] = min(latencies)
            result['max'] = max(latencies)
            latencies = sorted(latencies)
            result['p90'] = percentile(latencies, 0.90)
            result['p95'] = percentile(latencies, 0.95)
            result['p99'] = percentile(latencies, 0.99)
            result['p999'] = percentile(latencies, 0.999)
            results[key] = result

        headers = ['method', 'count', 'avg', 'min', 'max', 'p90', 'p95', 'p99', 'p999']
        data = []
        results = sorted(results.items(), key=lambda it: it[1][sort_by], reverse=True)

        def row(key, res):
            data = [key] + [res[header] for header in headers[1:]]
            return tuple(data)

        data = [row(key, result) for key, result in results]

        self._output.write('%s\n' % tabulate(data, headers=headers))
        self._output.flush()



# ==================================================================================================
# Copyright 2014 Twitter, Inc.
# --------------------------------------------------------------------------------------------------
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this work except in compliance with the License.
# You may obtain a copy of the License in the LICENSE file, or at:
#
#  http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==================================================================================================


""" network packets & header processing stuff

This comes from:
  https://github.com/twitter/zktraffic/blob/master/zktraffic/base/network.py

with a few corner cases handled (i.e.: header.unpack() might raise an exception)
"""

import socket

import dpkt


_loopback = dpkt.loopback.Loopback()
_ethernet = dpkt.ethernet.Ethernet()

def get_ip_packet(data, client_port, server_port, is_loopback=False):
    """ if client_port is 0 any client_port is good """
    header = _loopback if is_loopback else _ethernet

    try:
        header.unpack(data)
    except Exception as ex:
        raise ValueError('Bad header: %s' % ex)

    tcp_p = getattr(header.data, 'data', None)
    if type(tcp_p) != dpkt.tcp.TCP:
        raise ValueError('Not a TCP packet')

    if tcp_p.dport == server_port:
        if client_port != 0 and tcp_p.sport != client_port:
            raise ValueError('Request from different client')
    elif tcp_p.sport == server_port:
        if client_port != 0 and tcp_p.dport != client_port:
            raise ValueError('Reply for different client')
    else:
        raise ValueError('Packet not for/from client/server')

    return header.data


def get_ip(ip_packet, packed_addr):
    af_type = socket.AF_INET if type(ip_packet) == dpkt.ip.IP else socket.AF_INET6
    return socket.inet_ntop(af_type, packed_addr)



# -*- coding: utf-8 -*-

from __future__ import print_function

import argparse
import signal
import sys

from thrift.protocol.TBinaryProtocol import TBinaryProtocol
from thrift.protocol.TCompactProtocol import TCompactProtocol
from thrift.protocol.TJSONProtocol import TJSONProtocol

from .message_sniffer import MessageSnifferOptions, MessageSniffer
from .printer import FormatOptions, LatencyPrinter, PairedPrinter, Printer


VALID_PROTOCOLS = 'auto, binary, compact or json'


def get_flags():
    p = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    # general options
    p.add_argument('--iface', type=str, default='eth0', metavar='<iface>',
                   help='The interface to sniff from')
    p.add_argument('--port', type=int, default=9090, metavar='<port>',
                   help='The port the Thrift service listens too')
    p.add_argument('--max-queued', type=int, default=20*1024,
                   metavar='<maxqueued>',
                   help='Max number of queued messages')
    p.add_argument('--max-message-size', type=int, default=10*1024,
                   help='Max bytes size for a Thrift message')
    p.add_argument('--ip', type=str, nargs='+',
                   help='Only show messages from/to this IP(s)')
    p.add_argument('--finagle-thrift', default=False, action='store_true',
                   help='Detect finagle-thrift traffic (i.e.: with '
                   'request headers)')
    p.add_argument('--pcap-file', type=str, default='',
                   help='Path to pcap file, for offline introspection')
    p.add_argument('--debug', default=False, action='store_true',
                   help='Display debugging messages')
    p.add_argument('--protocol', type=str, default='auto',
                   help='Use a specific protocol. Options: %s' %
                   VALID_PROTOCOLS)

    cmds = p.add_subparsers(dest='cmd')

    # dump
    dump = cmds.add_parser('dump')
    dump.add_argument('--pretty', default=False, action='store_true',
                      help='Pretty print the Thrift structs')
    dump.add_argument('--color', default=False, action='store_true',
                      help='Use a different color for each stream')
    dump.add_argument('--unpaired', default=False, action='store_true',
                      help='Print (requests, replies) as they arrive, '
                      'possibly out of order')
    dump.add_argument('--show-header', default=False, action='store_true',
                      help='Show the header of each message, if any.'
                      'This makes sense with --finagle-thrift, given that '
                      'Finagle uses a RequestHeader')
    dump.add_argument('--show-fields', default=False, action='store_true',
                      help='Show the message\'s field')
    dump.add_argument('--show-values', default=False, action='store_true',
                      help='Show the values of each Thrift data type')
    dump.add_argument('--show-all', default=False, action='store_true',
                      help='Shows header, fields and values')
    dump.add_argument('--json', default=False, action='store_true',
                      help='Outputs messages as JSON')

    # stats
    stats = cmds.add_parser('stats')
    stats.add_argument('--count', type=int, default=0, metavar='<count>',
                       help='Waits for <count> pairs of (request, reply) and '
                       'reports the avg, p95 and p99 latencies')

    return p.parse_args()


def main():
    flags = get_flags()

    # route to the appropriate command
    if flags.cmd == 'stats':
        printer = LatencyPrinter(expected_calls=flags.count)
        read_values = False
    elif flags.cmd == 'dump':
        printer_cls = Printer if flags.unpaired else PairedPrinter
        format_opts = FormatOptions(
            flags.pretty,
            flags.color,
            flags.show_header or flags.show_all,
            flags.show_fields or flags.show_all,
            flags.json,
            )
        printer = printer_cls(format_opts)
        read_values = flags.show_values or flags.show_all
    else:
        print('Unknown command: %s' % flags.cmd)
        sys.exit(1)

    # which protocol to use
    if flags.protocol == 'auto':
        protocol = None
    elif flags.protocol == 'binary':
        protocol = TBinaryProtocol
    elif flags.protocol == 'compact':
        protocol = TCompactProtocol
    elif flags.protocol == 'json':
        protocol = TJSONProtocol
    else:
        print('Unknown protocol: %s' % flags.protocol)
        print('Valid options for --protocol are: %s' % VALID_PROTOCOLS)
        sys.exit(1)

    # launch the thrift message sniffer
    options = MessageSnifferOptions(
        iface=flags.iface,
        port=flags.port,
        ip=flags.ip,
        pcap_file=flags.pcap_file,
        protocol=protocol,
        finagle_thrift=flags.finagle_thrift,
        read_values=read_values,
        max_queued=flags.max_queued,
        max_message_size=flags.max_message_size,
        debug=flags.debug
        )
    message_sniffer = MessageSniffer(options, printer)

    def sigusr_handler(*args):
        print('message sniffer status: %s' % message_sniffer.status(),
              file=sys.stderr)

    # wire up sigusr1 for debugging info
    signal.signal(signal.SIGUSR1, sigusr_handler)

    # loop forever
    try:
        while message_sniffer.isAlive():
            message_sniffer.join(1)
    except (KeyboardInterrupt, SystemExit):
        pass
    finally:
        # force a stop
        message_sniffer.stop(wait_for_stopped=True)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

from __future__ import print_function

import argparse
import pprint
import sys

try:
    import mmap
    HAS_MMAP = True
except ImportError:
    HAS_MMAP = False

from .thrift_message import ThriftMessage


def get_flags():
    p = argparse.ArgumentParser(
        formatter_class=argparse.ArgumentDefaultsHelpFormatter)

    p.add_argument('file', type=str, default='-',
                   help='File from which to read Thrift messages')
    p.add_argument('--pretty', default=False, action='store_true',
                   help='Pretty print each Thrift message')
    p.add_argument('--finagle-thrift', default=False, action='store_true',
                   help='Detect finagle-thrift traffic (i.e.: with '
                   'request headers)')
    p.add_argument('--max-messages', type=int, default=-1, metavar='<max>',
                   help='Read up to <max> messages and then exit')
    p.add_argument('--skip-values', default=False, action='store_true',
                   help='Skip the values when reading each message (faster)')
    p.add_argument('--show-holes', default=False, action='store_true',
                   help='Show the holes in between messages (if any)')
    p.add_argument('--padding', type=int, default=0,
                   help='Number of bytes in between each message. For example, '
                   'if you using a binary log format that prepends a timestamp '
                   '(long) + the message length (int), the padding would be of '
                   '12 bytes')
    p.add_argument('--debug', default=False, action='store_true',
                   help='Display debugging messages')

    return p.parse_args()


def main():
    flags = get_flags()

    if flags.file == '-':
        fh = sys.stdin
    else:
        try:
            fh = open(flags.file)
        except IOError as ex:
            print('Couldn\'t open %s: %s' % (flags.file, ex))
            sys.exit(1)

    if HAS_MMAP and flags.file != '-':
        data = mmap.mmap(fh.fileno(), 0, access=mmap.ACCESS_READ)
        view = None
    else:
        # this might hurt...
        data = fh.read()
        view = memoryview(data)

    pp = pprint.PrettyPrinter(indent=4)

    start = flags.padding
    nread = 0

    # each hole is a tuple of (start_position, nbytes)
    holes = []

    try:
        done = False
        while start < len(data) and not done:
            for idx in range(start, len(data)):
                try:
                    msg, msglen = ThriftMessage.read(
                        view[idx:].tobytes() if view else data[idx:],
                        finagle_thrift=flags.finagle_thrift,
                        read_values=not flags.skip_values)

                    print(pp.pformat(msg.as_dict) if flags.pretty else msg)
                    nread += 1

                    # did we skip any bytes?
                    nbytes = idx - start
                    if nbytes > 0:
                        holes.append((start, nbytes))

                    start = idx + msglen + flags.padding

                    if flags.max_messages > 0 and nread > flags.max_messages:
                        done = True

                    break

                except Exception as ex:
                    if flags.debug:
                        print('Bad message: %s (idx=%d)' % (ex, idx))

    except KeyboardInterrupt:
        pass

    if holes:
        print('Read msgs: %d\nHoles: %d\n' % (nread, len(holes)))
        if flags.show_holes:
            for idx, hole in enumerate(holes, start=1):
                print('#%d: start=%d, size=%d' % (idx, hole[0], hole[1]))
    else:
        print('Read msgs: %d\nNo bytes skipped' % nread)



from setuptools import find_packages, setup

import os
import sys


PYTHON3 = sys.version_info > (3, )
HERE = os.path.abspath(os.path.dirname(__file__))


def readme():
    with open(os.path.join(HERE, 'README.rst')) as f:
        return f.read()


def get_version():
    with open(os.path.join(HERE, 'thrift_tools/__init__.py'), 'r') as f:
        content = ''.join(f.readlines())
    env = {}
    if PYTHON3:
        exec(content, env, env)
    else:
        compiled = compile(content, 'get_version', 'single')
        eval(compiled, env, env)
    return env['__version__']


setup(
    name='thrift-tools',
    version=get_version(),
    description='Thrift protocol analyzer',
    long_description=readme(),
    classifiers=[
        'Development Status :: 3 - Alpha',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2.7',
        'Topic :: System :: Distributed Computing',
        'Topic :: System :: Networking',
        ],
    keywords='thrift pcap',
    url='https://github.com/pinterest/thrift-tools',
    author='Raul Gutierrez Segales',
    author_email='rgs@pinterest.com',
    license='Apache',
    packages=find_packages(),
    test_suite='thrift_tools.tests',
    scripts=['bin/thrift-tool', 'bin/thrift-file-reader'],
      install_requires=[
          'ansicolors',
          'dpkt',
          'scapy==2.3.1',
          'thrift==0.9.2',
          'tabulate',
      ],
      tests_require=[
          'ansicolors',
          'dpkt',
          'nose',
          'scapy==2.3.1',
          'thrift==0.9.2',
          'tabulate',
      ],
      extras_require={
          'test': [
              'ansicolors',
              'dpkt',
              'nose',
              'scapy==2.3.1',
              'thrift==0.9.2',
              'tabulate',
              ],
      },
      include_package_data=True,
      zip_safe=False
)



from thrift.Thrift import TType


class ThriftStruct(object):
    """A thrift struct"""

    def __init__(self, fields):
        self._fields = fields

    @property
    def fields(self):
        return self._fields

    def is_isomorphic_to(self, other):
        """
        Returns true if all fields of other struct are isomorphic to this
        struct's fields
        """
        return (isinstance(other, self.__class__)
                and
                len(self.fields) == len(other.fields)
                and
                all(a.is_isomorphic_to(b) for a, b in zip(self.fields,
                                                          other.fields)))

    def __eq__(self, other):
        return (isinstance(other, self.__class__)
                and self.__dict__ == other.__dict__)

    def __len__(self):
        return len(self._fields)

    def __getitem__(self, key):
        return self._fields[key]

    def __iter__(self):
        return iter(self._fields)

    def __repr__(self):
        return "fields=%s" % self.fields

    class ObjectTooBig(Exception):
        pass

    @classmethod
    def read(cls,
             proto,
             max_fields,
             max_list_size,
             max_map_size,
             max_set_size,
             read_values=False):
        fields = []
        nfields = 0
        proto.readStructBegin()
        while True:
            nfields += 1
            if nfields >= max_fields:
                raise cls.ObjectTooBig('too many fields: %d' % nfields)

            _, ftype, fid = proto.readFieldBegin()
            if ftype == TType.STOP:
                break

            value = cls.read_field_value(
                proto, ftype,
                max_fields,
                max_list_size,
                max_map_size,
                max_set_size,
                read_values)

            proto.readFieldEnd()

            fields.append(ThriftField(cls.field_type_to_str(ftype), fid, value))
        proto.readStructEnd()
        return cls(fields)

    @classmethod
    def read_field_value(cls, proto, ftype,
                         max_fields,
                         max_list_size,
                         max_map_size,
                         max_set_size,
                         read_values):
        value = None

        def _read(_type):
            return cls.read_field_value(
                proto,
                _type,
                max_fields,
                max_list_size,
                max_map_size,
                max_set_size,
                read_values)

        # ##
        # There's a lot going on here:
        #
        # * for known scalar types, we check if we want the value or we skip
        # * for known collections, ditto but with sane size/limit checks
        #  * for the rest we skip
        #
        # Touching a line here should warrant writing another test case :-)
        #
        # FIXME: the way bytes are skipped is very lame, we should calculate
        #        the total number of bytes that are to be skipped, and have
        #        the transport seek() to that point.

        if ftype == TType.STRUCT:
            value = cls.read(
                proto,
                max_fields,
                max_list_size,
                max_map_size,
                max_set_size,
                read_values
            )
        elif ftype == TType.I32:
            if read_values:
                value = proto.readI32()
            else:
                proto.skip(ftype)
        elif ftype == TType.I64:
            if read_values:
                value = proto.readI64()
            else:
                proto.skip(ftype)
        elif ftype == TType.STRING:
            if read_values:
                value = proto.readString()
            else:
                proto.skip(ftype)
        elif ftype == TType.LIST:
            (etype, size) = proto.readListBegin()
            if size > max_list_size:
                raise cls.ObjectTooBig('list too long: %d' % size)
            value = []
            if read_values:
                value = [_read(etype) for _ in xrange(size)]
            else:
                for i in xrange(size):
                    proto.skip(etype)
            proto.readListEnd()
        elif ftype == TType.MAP:
            (ktype, vtype, size) = proto.readMapBegin()
            if size > max_map_size:
                raise cls.ObjectTooBig('map too big: %d' % size)
            value = {}
            if read_values:
                for i in xrange(size):
                    k = _read(ktype)
                    v = _read(vtype)
                    value[k] = v
            else:
                for i in xrange(size):
                    proto.skip(ktype)
                    proto.skip(vtype)
            proto.readMapEnd()
        elif ftype == TType.SET:
            (etype, size) = proto.readSetBegin()
            if size > max_set_size:
                raise cls.ObjectTooBig('set too big: %d' % size)
            value = set()
            if read_values:
                for i in xrange(size):
                    value.add(_read(etype))
            else:
                for i in xrange(size):
                    proto.skip(etype)
            proto.readSetEnd()
        else:
            # for now, we ignore all other values
            proto.skip(ftype)

        return value

    @staticmethod
    def field_type_to_str(ftype):
        if ftype == TType.STOP:
            return 'stop'
        elif ftype == TType.VOID:
            return 'void'
        elif ftype == TType.BOOL:
            return 'bool'
        elif ftype == TType.BYTE:
            return 'byte'
        elif ftype == TType.I08:
            return 'i08'
        elif ftype == TType.DOUBLE:
            return 'double'
        elif ftype == TType.I16:
            return 'i16'
        elif ftype == TType.I32:
            return 'i32'
        elif ftype == TType.I64:
            return 'i64'
        elif ftype == TType.STRING:
            return 'string'
        elif ftype == TType.UTF7:
            return 'utf7'
        elif ftype == TType.STRUCT:
            return 'struct'
        elif ftype == TType.MAP:
            return 'map'
        elif ftype == TType.SET:
            return 'set'
        elif ftype == TType.LIST:
            return 'list'
        elif ftype == TType.UTF8:
            return 'utf8'
        elif ftype == TType.UTF16:
            return 'utf16'
        else:
            raise ValueError('Unknown type: %s' % ftype)


class ThriftField(object):
    """A thrift field"""

    def __init__(self, field_type, field_id, value):
        self._field_type = field_type
        self._field_id = field_id
        self._value = value

    def __eq__(self, other):
        return (isinstance(other, self.__class__)
                and self.__dict__ == other.__dict__)

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self):
        return '(%s, %s, %s)' % (
            self.field_type, self.field_id, self._value)

    def is_isomorphic_to(self, other):
        """
        Returns true if other field's meta data (everything except value)
        is same as this one
        """
        return (isinstance(other, self.__class__)
                and self.field_type == other.field_type
                and self.field_id == other.field_id)

    @property
    def field_type(self):
        return self._field_type

    @property
    def field_id(self):
        return self._field_id

    @property
    def value(self):
        return self._value


import unittest

from thrift_tools.thrift_diff import ThriftDiff as Diff
from thrift_tools.thrift_message import ThriftMessage
from thrift_tools.thrift_struct import ThriftStruct, ThriftField


class ThriftDiffTestCase(unittest.TestCase):
    def test_is_diff_compatible(self):
        def _diff_compatibility(msg1, msg2, allowed):
            ok_to_diff, reason = Diff.can_diff(msg1, msg2)
            if allowed:
                self.assertTrue(ok_to_diff, reason)
            else:
                self.assertFalse(ok_to_diff)
                self.assertIsNotNone(reason)

        # Method name doesn't match
        rpc1 = ThriftMessage("ping", None, None, ThriftStruct([]))
        rpc2 = ThriftMessage("pong", None, None, ThriftStruct([]))
        _diff_compatibility(rpc1, rpc2, False)

        # Method names match, so does argument signature
        rpc1 = ThriftMessage("ping", None, None,
                             ThriftStruct([ThriftField("string", 1, "a")]))
        rpc2 = ThriftMessage("ping", None, None,
                             ThriftStruct([ThriftField("string", 1, "b")]))
        _diff_compatibility(rpc1, rpc2, True)

        # Method names match, but argument signature is different
        rpc1 = ThriftMessage("ping", None, None,
                             ThriftStruct([ThriftField("string", 1, "a")]))
        rpc2 = ThriftMessage("ping", None, None,
                             ThriftStruct([ThriftField("i32", 1, "b")]))
        _diff_compatibility(rpc1, rpc2, False)

    def test_diff_of_structs(self):
        f1 = ThriftField("string", 1, "one")
        f2 = ThriftField("i32", 2, 2)
        f3 = ThriftField("bool", 3, True)
        f4 = ThriftField("i32", 2, 4)
        s1 = ThriftStruct([f1, f2, f3])
        s2 = ThriftStruct([f1, f4])

        t_diff = Diff.of_structs(s1, s2)
        self.assertListEqual([(f1, f1), (f2, f4)], t_diff.common_fields)
        self.assertListEqual([f1], t_diff.fields_with_same_value)
        self.assertEqual(1, len(t_diff._fields_with_different_value))
        self.assertListEqual([f3], t_diff.fields_only_in_a)
        self.assertTrue(len(t_diff.fields_only_in_b) == 0)

    def test_diff_of_messages(self):
        f1 = ThriftField("string", 1, "one")
        f2 = ThriftField("i32", 2, 2)
        s1 = ThriftStruct([f1])
        s2 = ThriftStruct([f2])
        m1 = ThriftMessage("ping", "call", 1, s1)
        m2 = ThriftMessage("ping", "call", 2, s2)
        with self.assertRaisesRegexp(ValueError, 'argument signature'):
            Diff.of_messages(m1, m2)

        # Diff messages with all primitive args
        self.assertEqual(0, len(Diff.of_messages(m1, m1)),
                         "Message based diff only consider args of type struct")

        # Diff messages with mixed type of args
        f3 = ThriftField("struct", 1, ThriftStruct([f1, f2]))
        m3 = ThriftMessage("ping", "call", 2, ThriftStruct([f1, f2, f3, f3]))
        diffs = Diff.of_messages(m3, m3)
        self.assertEqual(2, len(diffs))

        # We diffed identical messages, verify that diff reflects that
        self.assertListEqual([f1, f2], diffs[0].fields_with_same_value)
        self.assertEqual(0, len(diffs[0].field_with_different_value))





import os

_resources_dir = os.path.join(
    os.path.realpath(os.path.dirname(__file__)),
    'resources'
)


def get_pcap_path(capture_file):
    return os.path.join(_resources_dir, '%s.pcap' % (capture_file))
